/* FILE:        BFsimpleGLUT.cpp
 * DATE:        5/8/2014
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2014, BitFlow, Inc.
 * DESCRIPTION: A simplified interface for displaying a single grayscale image using FreeGLUT.
 */

#define NOMINMAX

#include "BFsimpleGLUT.h"
#include "SimpleTimers.hpp"

// Standard includes.
#include <limits>
#include <cmath>

#include <atomic>
#include <mutex>

#include <string>

// OpenGL includes (for display).
#include <GL\glew.h>    
#ifdef _WIN32
#include <GL\wglew.h>
#endif

// GLUT includes.
#include <GL\freeglut.h>

// The OpenGL display variables and function.
struct GLBufferParams
{
    GLuint                  tex;
    cl_mem                  texCL;

    cl_context              oclContext;
    GLint                   width;
    GLint                   height;
    GLint                   bytedepth;
    size_t                  size;

    GLBufferParams (void)
        : tex           (0)
        , texCL         (0)
        , oclContext    (0)
        , width         (0)
        , height        (0)
        , bytedepth     (0)
        , size          (0)
    { }

    void clean (void)
    {
        if (tex)
            glDeleteTextures(1, &tex);
        
        tex = 0;
        texCL = 0;

        oclContext = 0;
        width = 0;
        height = 0;
        bytedepth = 0;
        size = 0;
    }

    bool make (const cl_context context, const int w, const int h, const int b)
    {
        clean();

        oclContext = context;
        width = w;
        height = h;
        bytedepth = b;
        size = width * height * bytedepth;

        // allocate new buffers
        glEnable(GL_TEXTURE_2D);
        glGenTextures(1, &tex);
        glBindTexture(GL_TEXTURE_2D, tex);
        
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,     GL_CLAMP_TO_EDGE);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,     GL_CLAMP_TO_EDGE);

        if (1 == bytedepth || 2 == bytedepth)
        {
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_SWIZZLE_G_EXT, GL_RED);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_SWIZZLE_B_EXT, GL_RED);
        }

        if (1 == bytedepth)
            glTexImage2D(GL_TEXTURE_2D, 0, GL_R8, width, height, 0, GL_RED, GL_UNSIGNED_BYTE, NULL);
        else if (2 == bytedepth)
            glTexImage2D(GL_TEXTURE_2D, 0, GL_R16, width, height, 0, GL_RED, GL_UNSIGNED_SHORT, NULL);
        else if (3 == bytedepth)
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
        
        cl_int oclrc;
        texCL = clCreateFromGLTexture(oclContext, CL_MEM_WRITE_ONLY, GL_TEXTURE_2D, 0, tex, &oclrc);
        if (CL_SUCCESS != oclrc)
        {
            clean();
            return false;
        }

        return true;
    }
};

// BFsimpleGLUT global parameters.
GLBufferParams          dispBuffer;
std::atomic<cl_mem>     oclBufferAtm (0);

    // Externally accessible variables.
std::string             overlayText;
RateAverager<>          bufferDisplayRate;
std::mutex              externallyAccessibleVariableMutex;

BFsimpleGLUTkeyCallback keyCallback = nullptr;

GLfloat dispWidth, dispHeight;
GLfloat zoomScalar = 1.f;
GLint xOffset = 0, yOffset = 0;

bool smoothTransitionsEnabled = true;

// Callback to time display updates.
void gfsdTimerEventFunc (int value)
{
    // Continue updating the display at about 60Hz.
    glutPostRedisplay();
    glutTimerFunc(1000/60, gfsdTimerEventFunc, 0);
}

// Callback to update the viewport.
void gfsdReshapeFunc (int width, int height)
{
    dispWidth = GLfloat(width);
    dispHeight = GLfloat(height);
    glViewport(0, 0, width, height);
}

// Smoothly transition the value of a.
float transitionToValue (float &a, const float d, const float by)
{
    const float aInit = a;
    const float absolute = abs(a - d);
    const float maximum = absolute * by;

    if (!smoothTransitionsEnabled)
        a = d;
    else if (a > d)
        a -= std::min(maximum, absolute);
    else if (d > a)
        a += std::min(maximum, absolute);

    return a - aInit;
}

void renderText (std::string const& text, const int lineWidth, const int x, const int y, const float r, const float g, const float b, const float a = 1.f)
{
    glPushMatrix();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluOrtho2D(0.0, dispWidth, 0.0, dispHeight);

    glDisable(GL_TEXTURE_2D);
    
    glTranslatef(float(x), float(y), 0);
    glScalef(0.2f, 0.2f, 0.2f);
    
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor4f(r, g, b, a);
    glLineWidth((GLfloat)lineWidth);
    
    glutStrokeString( GLUT_STROKE_MONO_ROMAN, (unsigned char*)text.c_str() );
    
    glDisable(GL_BLEND);
    
    glEnable(GL_TEXTURE_2D);
        
    glPopMatrix();
}

// Callback to display a buffer.
GLfloat displayScaling = 1.f, displayXOffset = 0.f, displayYOffset = 0.f;
void gfsdDisplayFunc (void)
{
    //// Setup the display area.
    
    // Reset the coordinate system.
    glLoadIdentity();

    // Clear the display to a light blue color.
    glClearColor(0.8f, 1.f, 1.f, 0.f);
    glClear(GL_COLOR_BUFFER_BIT);

    glPushMatrix();
    glColor3f(1, 1, 1);
    {
        //// Adjust the display.
    
        // Pan first, then scale.
        const GLfloat xTransMag = abs(transitionToValue(displayXOffset, GLfloat(xOffset), 0.1f));
        const GLfloat yTransMag = abs(transitionToValue(displayYOffset, GLfloat(yOffset), 0.1f));
        if (0.2f > xTransMag && 0.2f > yTransMag)
            transitionToValue(displayScaling, zoomScalar, 0.1f);
    
        // Set the scaling.
        glScalef(displayScaling, displayScaling, 1.f);
    
        // Set the offset.
        const float xOffsetScalar = -2.f * displayXOffset / float(dispWidth);
        const float yOffsetScalar = -2.f * displayYOffset / float(dispHeight);
        glTranslatef(xOffsetScalar, yOffsetScalar, 0.f);
    
        //// Display the main buffer.
    
        // load texture
        glBindTexture(GL_TEXTURE_2D, dispBuffer.tex);
    
        // Determine the "real" display coordinates.
        const GLfloat bufWidth = GLfloat(dispBuffer.width);
        const GLfloat bufHeight = GLfloat(dispBuffer.height);
    
        const GLfloat xScalar = bufWidth / dispWidth;
        const GLfloat yScalar = bufHeight / dispHeight;
    
        const GLfloat upper_left[]  = { -1.f * xScalar, 1.f * yScalar };
        const GLfloat lower_left[]  = { -1.f * xScalar, -1.f * yScalar };
        const GLfloat upper_right[] = { 1.f * xScalar,  1.f * yScalar };
        const GLfloat lower_right[] = { 1.f * xScalar,  -1.f * yScalar };
    
        // Setup the texture quadrant.
        glBegin(GL_QUADS);
        glTexCoord2f(0.0f, 0.0f);   glVertex2fv(upper_left);
        glTexCoord2f(0.0f, 1.0f);   glVertex2fv(lower_left);
        glTexCoord2f(1.0f, 1.0f);   glVertex2fv(lower_right);
        glTexCoord2f(1.0f, 0.0f);   glVertex2fv(upper_right);
        glEnd();
    }
    glPopMatrix();
    
    //// Display the overlay text.
    const auto overlayTextCommitted = [](void)
    {
        const std::lock_guard<std::mutex> l (externallyAccessibleVariableMutex);
        return overlayText;
    }();

    if (overlayTextCommitted.size())
    {
        renderText(overlayTextCommitted, 2, 24, int(dispHeight) - 24, 0, 0, 0);
        renderText(overlayTextCommitted, 2, 25, int(dispHeight) - 25, 1, 1, 1);
    }

    //// Update the display surface.
    
    // Count a displayed buffer.
    {
        const std::lock_guard<std::mutex> l (externallyAccessibleVariableMutex);
        bufferDisplayRate.count();
    }

    // Swap the front and back buffers.
    glutSwapBuffers();
}

// Keyboard handlers.
void gfsdSpecialFunc (int k, int, int)
{
    switch (k)
    {
    case GLUT_KEY_F11:
        glutFullScreenToggle();
        break;

    case GLUT_KEY_UP:
        yOffset += 10;
        break;
    case GLUT_KEY_DOWN:
        yOffset -= 10;
        break;
    case GLUT_KEY_LEFT:
        xOffset -= 10;
        break;
    case GLUT_KEY_RIGHT:
        xOffset += 10;
        break;
    }
}
void gfsdKeyboardFunc (unsigned char k, int, int)
{
    if (keyCallback && keyCallback(k))
        return;

    switch (k)
    {
    case 'Q': case 'q':
    case VK_ESCAPE:
        glutLeaveMainLoop();
        break;

    case '-':
        zoomScalar = std::max( pow(1.1f, -50.f), zoomScalar / 1.1f);
        break;
    case '+':
        zoomScalar = std::min( pow(1.1f, 50.f), zoomScalar * 1.1f);
        break;

    case '1':
        // Zoom to 100%, and nullify the offsets.
        xOffset = 0;
        yOffset = 0;
        zoomScalar = 1.f;
        break;

    case 'F': case 'f':
        // Zoom to fit the display area, and nullify the offsets.
        xOffset = 0;
        yOffset = 0;

        // Select the most appropriate scaling.
        zoomScalar = std::min(dispWidth / GLfloat(dispBuffer.width), dispHeight / GLfloat(dispBuffer.height));
        zoomScalar = std::min( pow(1.1f, 50.f), std::max( pow(1.1f, -50.f), zoomScalar ) );
        break;
    }
}

// Mouse wheel handler.
void gfsdMouseWheelFunc (int wheel, int direction, int x, int y)
{
    zoomScalar = std::max(std::min( zoomScalar * pow(1.1f, (float)direction), pow(1.1f, 50.f) ), pow(1.1f, -50.f));
}

// Mouse button handler.
bool panDrag = false;
int xOffsetInit = 0, yOffsetInit = 0;
int xPosInit = 0, yPosInit = 0;
void gfsdMouseFunc (int button, int state, int x, int y)
{
    if (GLUT_DOWN == state)
    {
        // Enable the appropriate drag operation.
        panDrag = (GLUT_LEFT_BUTTON == button);

        xOffsetInit = xOffset;
        yOffsetInit = yOffset;
        xPosInit = x;
        yPosInit = y;

        smoothTransitionsEnabled = false;
    }
    else
        smoothTransitionsEnabled = true;
}

// Mouse drag handler.
void gfsdMotionFunc (int x, int y)
{
    const float xDelta = float(xPosInit - x);
    const float yDelta = float(y - yPosInit);
    if (abs(xDelta) < 1.f || abs(yDelta) < 1.f)
        return;
    
    if (panDrag)
    {
        xOffset = (xOffsetInit + int(xDelta / zoomScalar));
        yOffset = (yOffsetInit + int(yDelta / zoomScalar));
    }
}

// GLUT configurators.
bool initOpenGLBuffers (const cl_context oclContext, unsigned int width, unsigned int height, unsigned int bytedepth)
{
    // delete old buffers
    dispBuffer.clean();

    // check for minimized window
    if ((width == 0) && (height == 0))
        return true;

    // Allocate the buffers.
    return dispBuffer.make(oclContext, width, height, bytedepth);
}

#ifdef _WIN32
// This is specifically to enable the application to enable/disable vsync
typedef BOOL (WINAPI *PFNWGLSWAPINTERVALFARPROC)(int);

void setVSync (int interval)
{
    if (WGL_EXT_swap_control)
    {
        wglSwapIntervalEXT = (PFNWGLSWAPINTERVALFARPROC)wglGetProcAddress("wglSwapIntervalEXT");
        wglSwapIntervalEXT(interval);
    }
}
#endif

// Initialize BFsimpleGLUT. Call this once, prior to using any other methods.
bool initBFsimpleGLUT ( int argc, char *argv[] )
{
    // Setup the OpenGL window.
    glutInit(&argc, argv);
    
    glutInitDisplayMode(GLUT_LUMINANCE | GLUT_DOUBLE | GLUT_MULTISAMPLE);
    glutInitWindowSize(256, 256);
    glutInitWindowPosition(0, 0);
    glutCreateWindow(argv[0]);
    
    // Update the display at about 60Hz.
    glutTimerFunc(1000/60, gfsdTimerEventFunc, 0);
    glutDisplayFunc(gfsdDisplayFunc);
    glutReshapeFunc(gfsdReshapeFunc);

    glutSpecialFunc(gfsdSpecialFunc);
    glutKeyboardFunc(gfsdKeyboardFunc);
    glutMouseWheelFunc(gfsdMouseWheelFunc);
    glutMouseFunc(gfsdMouseFunc);
    glutMotionFunc(gfsdMotionFunc);

    glewInit();

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
    
#ifdef _WIN32
    setVSync(0);
#endif
    
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

    return true;
}

// Prepare the BFsimpleGLUT buffers. To be called after enterBFsimpleGLUT.
bool makeBFsimpleGLUTbuffers
    ( const cl_context oclContext
    , const unsigned int width
    , const unsigned int height
    , const unsigned int bitdepth )
{
    const unsigned int imagePixBytes = (bitdepth + 7) / 8;
    return initOpenGLBuffers(oclContext, width, height, imagePixBytes);
}

// Enter the main-loop of the BFsimpleGLUT engine. This will return when the user closes the display
// window. Only on thread should call this method at any given time.
void enterBFsimpleGLUT ( const bool fullScreen )
{
    if (fullScreen)
        glutFullScreen();

    // Reset the rate counter.
    {
        const std::lock_guard<std::mutex> externallyAccessibleVariableLock (externallyAccessibleVariableMutex);
        bufferDisplayRate.reset();
    }

    // The GL main loop.
    glutMainLoop();
}

// Exit the GLUT main-loop on demand, closing any windows.
void exitBFsimpleGLUT (void)
{
    glutLeaveMainLoop();
}

// Setup a callback function to be called whenever an ASCII keyboard key is pressed. If this function
// returns true, the display will ignore the keypress internally. If false, the keypress may invoke
// some standard internal behavior.
void setBFsimpleGLUTkeyCallback (BFsimpleGLUTkeyCallback callback)
{
    keyCallback = callback;
}

// Pass a buffer to BFsimpleGLUT for display (we target 60Hz display; some buffers may be dropped).
cl_int enqueueSetMainBFsimpleGLUT
    ( const cl_command_queue oclCmdQueue
    , const cl_mem oclBuffer
    , const cl_uint numEventsInWaitList
    , const cl_event *const eventWaitList
    , cl_event *oclEvent )
{
    cl_int oclrc;

    oclrc = clEnqueueAcquireGLObjects(oclCmdQueue, 1, &dispBuffer.texCL, 0, NULL, NULL);
    if (CL_SUCCESS != oclrc)
        return oclrc;

    size_t dstOrigin[] = { 0, 0, 0 };
    size_t region[] = { dispBuffer.width, dispBuffer.height, 1 };
    oclrc = clEnqueueCopyBufferToImage(oclCmdQueue, oclBuffer, dispBuffer.texCL, 0, dstOrigin, region, numEventsInWaitList, eventWaitList, oclEvent);
    if (CL_SUCCESS != oclrc)
        return oclrc;
    
    oclrc = clEnqueueReleaseGLObjects(oclCmdQueue, 1, &dispBuffer.texCL, 0, NULL, NULL);

    return oclrc;
}

// Set the overlay text.
void setBFsimpleGLUToverlayText (const char *const text)
{
    const std::lock_guard<std::mutex> externallyAccessibleVariableLock (externallyAccessibleVariableMutex);
    overlayText = text;
}

// Return the estimated rate in Hz at which frames are being displayed.
double getBFsimpleGLUTdisplayRateAverage (void)
{
    const std::lock_guard<std::mutex> externallyAccessibleVariableLock (externallyAccessibleVariableMutex);
    return bufferDisplayRate.average_rate();
}
