/* FILE:        SimpleDGMA.cpp
 * DATE:        6/9/2015
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2015, BitFlow, Inc.
 * DESCRIPTION: Example program demonstrating AMD DirectGMA with simple OpenCL processing
 *              and an optional OpenGL display of the results.
 */

#if !defined(NOMINMAX)
#    define NOMINMAX
#endif

#define USE_GL_DISPLAY (1)

// BitFlow includes
#include <BiApi.h>
#include <BFErApi.h>

#if USE_GL_DISPLAY
#   include "../BFsimpleGLUT/BFsimpleGLUT.h"
#endif

// OpenCL includes
#include <CL/opencl.h>

#define clGetGLContextInfoKHR   clGetGLContextInfoKHR_proc

// STD includes
#include <iostream>
#include <fstream>
#include <sstream>

#include <vector>
#include <thread>
#include <future>
#include <atomic>

#include <ctime>
#include <cmath>

int oclOffset = 0;
double oclGain = 1.;

bool keyHandler (const char key)
{
    switch (key)
    {
    case '[':
        oclGain /= 1.025;
        return true;
    case ']':
        oclGain *= 1.025;
        return true;
    case '\\':
        oclGain = 1.;

    case '{':
        --oclOffset;
        return true;
    case '}':
        ++oclOffset;
        return true;
    case '|':
        oclOffset = 0;
    }

    return false;
}

// Simple class to execute a queue of actions in its destructor.
class ScopedUnwinder
{
    typedef std::function<void(void)> Action;
    typedef std::vector<Action> ActionQueue;
    ActionQueue m_queue;

    ScopedUnwinder (ScopedUnwinder const&);
    ScopedUnwinder& operator= (ScopedUnwinder const&);

public:
    ScopedUnwinder (void)
    { }

    ~ScopedUnwinder (void)
    {
        for (auto I = m_queue.crbegin(), E = m_queue.crend(); I != E; I++)
            (*I)();
    }

    template<typename F>
    inline void spool (F &&f)
    {
        m_queue.push_back( std::forward<F>(f) );
    }
};

int main (int argc, char *argv[])
{
    // Variables for DirectGMA.
    clGetGLContextInfoKHR_fn            clGetGLContextInfoKHR           = NULL;
    clEnqueueWaitSignalAMD_fn           clEnqueueWaitSignalAMD          = NULL; 
    clEnqueueWriteSignalAMD_fn          clEnqueueWriteSignalAMD         = NULL;
    clEnqueueMakeBuffersResidentAMD_fn  clEnqueueMakeBuffersResidentAMD = NULL;

    // String constants.
    const std::string gpuVendor = "Advanced Micro Devices, Inc.";

    // BitFlow variables.
    BFRC rc;
    BFU32 BoardType, BoardNum;
    BFU32 Init, SerNum;
    Bd hBoard;
    BFU32 imageWidth, imageHeight, imageBitDepth, imageByteDepth, imageByteSize;
    BIBA hBiba;
    BFU32 bufCnt = 25;
    std::vector<BFU8> sysDispBuf;

    // OpenCL variables.
    cl_int clrc;
    cl_uint oclPlatformCnt;
    std::vector<cl_platform_id> oclPlatforms;
    cl_platform_id oclAmdPlatform;
    cl_device_type oclDevType = CL_DEVICE_TYPE_GPU;
    cl_context oclContext;
    size_t oclDeviceCnt;
    std::vector<cl_device_id> oclDevices;
    cl_device_id oclAmdDevice;
    cl_command_queue_properties oclCmdQueueProps = CL_QUEUE_PROFILING_ENABLE;
    cl_command_queue oclCmdQueue;
    std::vector<cl_mem> oclBufs;
    std::vector<cl_bus_address_amd> oclBusAddrs;
    cl_program oclProgram;
    cl_kernel oclKrnBufferGain;
    
    // Synchronization variables.
#if USE_GL_DISPLAY
    std::atomic<bool> glutHasExited = false;
#endif
    
    // Unwinder for the above variables, which it will reference, and therefore must follow.
    ScopedUnwinder unwinder;
    
    // Query the user for the BitFlow board to be opened.
    if ( IDCANCEL == DoBrdOpenDialog(BOD_HIDEJUSTOPEN, FF_BITFLOW_MODERN, &BoardType, &BoardNum, &Init, &SerNum) )
        return EXIT_SUCCESS;

    //// Configure the BitFlow board for acquisition.

    // Open the BitFlow board specified by the user.
    rc = BiBrdOpen(BoardType, BoardNum, &hBoard);
    if (BF_OK != rc)
    {
        std::cerr << "Failed to open BitFlow board " << BoardNum << ".\n";
        return EXIT_FAILURE;
    }
    unwinder.spool([&](){ BiBrdClose(hBoard); });

    // Retrieve the requisite board parameters.
    BiBrdInquire(hBoard, BiCamInqXSize, &imageWidth);
    BiBrdInquire(hBoard, BiCamInqYSize0, &imageHeight);
    BiBrdInquire(hBoard, CiCamInqPixBitDepth, &imageBitDepth);

    const BFU32 supportedDepths[] = { 8, 10, 12, 14, 16 };
    if (std::end(supportedDepths) == std::find(std::begin(supportedDepths), std::end(supportedDepths), imageBitDepth))
    {
        std::cerr << "Unsupported bit-depth, " << imageBitDepth << ".\n";
        return EXIT_FAILURE;
    }

    imageByteDepth = (imageBitDepth + 7) / 8;
    imageByteSize = imageWidth * imageHeight * imageByteDepth;

        // Setup a useful default gain.
    oclOffset = 0;
    if (8 == imageBitDepth)
        oclGain = 1.;
    else
        oclGain = (pow(2., 16.) - 1.) / (pow(2., imageBitDepth) - 1.);
    
    //// Configure an OpenCL device for processing.

    // Get the available OpenCL platforms.
        // Enumerate
    clrc = clGetPlatformIDs(0, NULL, &oclPlatformCnt);
    if (CL_SUCCESS != clrc)
    {
        std::cerr << "Failed to enumerate the available OpenCL platforms.\n";
        return EXIT_FAILURE;
    }

        // Validate
    if (0 >= oclPlatformCnt)
    {
        std::cerr << "There aren't any OpenCL platforms available on this system.\n";
        return EXIT_FAILURE;
    }

        // Query
    oclPlatforms.resize(oclPlatformCnt);
    clrc = clGetPlatformIDs(oclPlatformCnt, oclPlatforms.data(), &oclPlatformCnt);
    if (CL_SUCCESS != clrc)
    {
        std::cerr << "Unable to get the list of available OpenCL platforms.\n";
        return EXIT_FAILURE;
    }

        // Loop through platforms and check if the vendor is equal to AMD
    for (auto const& platform : oclPlatforms)
    {
        size_t propSize = 0;
        clrc = clGetPlatformInfo(platform, CL_PLATFORM_VENDOR, 0, NULL, &propSize);
        if (CL_SUCCESS != clrc)
            continue;
        
        std::vector<char> propVal (propSize);
        clrc = clGetPlatformInfo(platform, CL_PLATFORM_VENDOR, propSize, propVal.data(), &propSize);
        if (CL_SUCCESS != clrc)
            continue;
        
        if (gpuVendor == propVal.data())
        {
            oclAmdPlatform = platform;
            break;
        }
    }

    // Create an OpenCL context for GPU type devices.
#if USE_GL_DISPLAY

    // Register the GLUT key handler.
    setBFsimpleGLUTkeyCallback(keyHandler);

    // Enter the GLUT display thread, setting up the OpenCL context, as we go.
    std::promise<bool> glutEnterPromise;
    auto glutExitFuture = std::async( [&](void) -> bool
        {
            // Initialize BFsimpleGLUT.
            if (!initBFsimpleGLUT(argc, argv))
            {
                std::cerr << "Failed to initialize BFsimpleGLUT.\n";
                glutEnterPromise.set_value(false);
                return false;
            }

            // Create an OpenCL context for ths GLUT OpenGL context.
            cl_context_properties oclContextProps[] =
                {
                    CL_CONTEXT_PLATFORM,    (cl_context_properties)oclAmdPlatform,
#   ifdef _WIN32
                    CL_GL_CONTEXT_KHR,      (cl_context_properties)wglGetCurrentContext(),
                    CL_WGL_HDC_KHR,         (cl_context_properties)wglGetCurrentDC(),
#   else
#       error "WGL sharing TODO on this platform."
#   endif
                    0
                };
    
            oclContext = clCreateContextFromType(oclContextProps, oclDevType, NULL, NULL, &clrc);
            if (CL_SUCCESS != clrc)
            {
                std::cerr << "Failed to create the OpenCL context for the OpenGL context.\n";
                glutEnterPromise.set_value(false);
                return false;
            }
            unwinder.spool([oclContext](void){ clReleaseContext(oclContext); });

            // Initialize the GLUT buffers.
            if (!makeBFsimpleGLUTbuffers(oclContext, imageWidth, imageHeight, imageBitDepth))
            {
                std::cerr << "Failed the make the BFsimpleGLUT buffers.\n";
                glutEnterPromise.set_value(false);
                return false;
            }

            // Enter the GLUT main loop.
            glutEnterPromise.set_value(true);
            enterBFsimpleGLUT();
            glutHasExited.store(true);

            return true;
        } );

    if (!glutEnterPromise.get_future().get())
    {
        std::cerr << "A problem occured during GLUT initialization.\n";
        return EXIT_FAILURE;
    }

#else // USE_GL_DISPLAY

    cl_context_properties oclContextProps[] =
        {
            CL_CONTEXT_PLATFORM,    (cl_context_properties)oclAmdPlatform,
            0
        };
    oclContext = clCreateContextFromType(oclContextProps, oclDevType, NULL, NULL, &clrc);
    if (CL_SUCCESS != clrc)
    {
        std::cerr << "Failed to create the OpenCL context.\n";
        return EXIT_FAILURE;
    }
    unwinder.spool([&](){ clReleaseContext(oclContext); });
    

#endif // USE_GL_DISPLAY

    // Get the OpenCL devices for the context.
        // Enumerate
    clrc = clGetContextInfo(oclContext, CL_CONTEXT_DEVICES, 0, NULL, &oclDeviceCnt);
    if (CL_SUCCESS != clrc)
    {
        std::cerr << "Unable to get the number of OpenCL devices on this context.\n";
        return EXIT_FAILURE;
    }

        // Validate
    if (0 >= oclDeviceCnt)
    {
        std::cerr << "There aren't any devices on this OpenCL context.\n";
        return EXIT_FAILURE;
    }
    
        // Query
    oclDevices.resize(oclDeviceCnt);
    clrc = clGetContextInfo(oclContext, CL_CONTEXT_DEVICES, oclDeviceCnt, oclDevices.data(), &oclDeviceCnt);
    if (CL_SUCCESS != clrc)
    {
        std::cerr << "Unable to get the list of OpenCL devices on this context.\n";
        return EXIT_FAILURE;
    }

    // Find the first AMD device.
    oclAmdDevice = nullptr;
    for (auto devId : oclDevices)
    {
        size_t propSize = 0;
        clrc = clGetDeviceInfo(devId, CL_DEVICE_VENDOR, 0, NULL, &propSize);
        if (CL_SUCCESS != clrc)
            continue;
        
        std::vector<char> propVal (propSize);
        clrc = clGetDeviceInfo(devId, CL_DEVICE_VENDOR, propSize, propVal.data(), &propSize);
        if (CL_SUCCESS != clrc)
            continue;
        
        if (gpuVendor == propVal.data())
        {
            oclAmdDevice = devId;
            break;
        }
    }

    if (!oclAmdDevice)
    {
        std::cerr << "Unable to find an AMD device on this system.\n";
        return EXIT_FAILURE;
    }

    // Setup an OpenCL command queue.
    oclCmdQueue = clCreateCommandQueue(oclContext, oclAmdDevice, oclCmdQueueProps, &clrc);
    if (CL_SUCCESS != clrc)
    {
        std::cerr << "Failed to create the OpenCL command queue.\n";
        return EXIT_FAILURE;
    }
    unwinder.spool([&](){ clReleaseCommandQueue(oclCmdQueue); });
    
    // Load required extensions.
    clGetGLContextInfoKHR =             (clGetGLContextInfoKHR_fn)clGetExtensionFunctionAddressForPlatform(oclAmdPlatform, "clGetGLContextInfoKHR");
    clEnqueueWaitSignalAMD =            (clEnqueueWaitSignalAMD_fn)clGetExtensionFunctionAddressForPlatform(oclAmdPlatform, "clEnqueueWaitSignalAMD");
    clEnqueueWriteSignalAMD =           (clEnqueueWriteSignalAMD_fn)clGetExtensionFunctionAddressForPlatform(oclAmdPlatform, "clEnqueueWriteSignalAMD");
    clEnqueueMakeBuffersResidentAMD =   (clEnqueueMakeBuffersResidentAMD_fn)clGetExtensionFunctionAddressForPlatform(oclAmdPlatform, "clEnqueueMakeBuffersResidentAMD");

    if (!clGetGLContextInfoKHR || !clEnqueueWaitSignalAMD || !clEnqueueWriteSignalAMD || !clEnqueueMakeBuffersResidentAMD)
    {
        std::cerr << "Failed to load the DirectGMA extension methods.\n";
        return EXIT_FAILURE;
    }

    // Create the OpenCL buffers.
    oclBufs.resize(bufCnt, 0);
    for (auto &buf : oclBufs)
    {
        buf = clCreateBuffer(oclContext, CL_MEM_BUS_ADDRESSABLE_AMD, imageByteSize, NULL, &clrc);
        if (CL_SUCCESS != clrc)
        {
            std::cerr << "Failed to create the processing buffer.\n";
            return EXIT_FAILURE;
        }
        unwinder.spool([buf](){ clReleaseMemObject(buf); });
    }

    oclBusAddrs.resize(bufCnt);
    memset(oclBusAddrs.data(), 0, oclBusAddrs.size() * sizeof(cl_bus_address_amd));

    // Make the OpenCL buffers resident.
    clrc = clEnqueueMakeBuffersResidentAMD(oclCmdQueue, bufCnt, oclBufs.data(), true, oclBusAddrs.data(), 0, 0, 0);
    if (CL_SUCCESS != clrc)
    {
        std::cerr << "Failed to make the OpenCL buffers resident for DirectGMA.\n";
        return EXIT_FAILURE;
    }

    // Assign BitFlow acquisition buffers.
    std::vector<PBFU32> oclBufAddrs;
    for (auto const& busAddr : oclBusAddrs)
        oclBufAddrs.push_back( (PBFU32)busAddr.surface_bus_address );

    rc = BiBufferAssign(hBoard, &hBiba, oclBufAddrs.data(), bufCnt);
    if (BF_OK != rc)
    {
        std::cerr << "Failed to allocate " << bufCnt << " system acquisition buffers.\n";
        return EXIT_FAILURE;
    }
    unwinder.spool([&](){ BiBufferUnassign(hBoard, &hBiba); });

    // Create a system return buffer.
    try
    {
        sysDispBuf.resize(imageByteSize);
    }
    catch (std::bad_alloc const& err)
    {
        std::cerr << "Failed to allocate the system display buffer: " << err.what() << "\n";
        return EXIT_FAILURE;
    }

    // Setup the BitFlow acquisition engine.
    rc = BiCircAqSetup(hBoard, &hBiba, CirErStop, BiAqEngJ|DestAddrIsPhys);
    if (BF_OK != rc)
    {
        do
        {
            BiErrorShow(hBoard, rc);
            rc = BiCirErrorCheck(hBoard, &hBiba);
        }
        while (BF_OK != rc);

        std::cerr << "Failed to allocate " << bufCnt << " system acquisition buffers.\n";
        return EXIT_FAILURE;
    }
    unwinder.spool([&](){ BiCircCleanUp(hBoard, &hBiba); });
    
    //// Setup the kernel.

    // Load the file.
    std::ifstream bufferGainFStrm (".\\BufferGain.cl");
    if (!bufferGainFStrm)
    {
        std::cerr << "Failed to open OpenCL kernel source.\n";
        return EXIT_FAILURE;
    }

    std::stringstream bufferGainSStrm;
    bufferGainSStrm << bufferGainFStrm.rdbuf();
    const std::string bufferGainStr = bufferGainSStrm.str();

    const char *kernelSrc[] = { bufferGainStr.c_str() };
    const size_t kernelSize[] = { bufferGainStr.size() };

    // Create the program.
    oclProgram = clCreateProgramWithSource(oclContext, 1, kernelSrc, kernelSize, &clrc);
    if (CL_SUCCESS != clrc)
    {
        std::cerr << "Failed to create OpenCL program from the kernel source.\n";
        return EXIT_FAILURE;
    }
    unwinder.spool([&](){ clReleaseProgram(oclProgram); });

    // Build the program.
    clrc = clBuildProgram(oclProgram, 1, &oclAmdDevice, "", NULL, NULL);
    if (CL_SUCCESS != clrc)
    {
        std::cerr << "Failed to build the OpenCL program.\n";
        return EXIT_FAILURE;
    }

    // Create the kernel.
    if (8 == imageBitDepth)
        oclKrnBufferGain = clCreateKernel(oclProgram, "buffer_gain_8bit", &clrc);
    else
        oclKrnBufferGain = clCreateKernel(oclProgram, "buffer_gain_16bit", &clrc);

    if (CL_SUCCESS != clrc)
    {
        std::cerr << "Failed to create the OpenCL kernel.\n";
        return EXIT_FAILURE;
    }
    unwinder.spool([&](){ clReleaseKernel(oclKrnBufferGain); });

    // Create a lambda to execute the kernel.
    const auto enqueue_buffer_gain = [oclCmdQueue,oclKrnBufferGain]
        (cl_mem dst, const cl_mem src, cl_uint bufSize, cl_uint byteDepth, const double gain, cl_int offset, cl_uint waitEvtCnt, cl_event waitEvts[], cl_event *kernelEvt)
        {
            cl_int qrc;
            unsigned int arg = 0;

            const int gainDenominator = (int)(pow(2., 8 * byteDepth) - 1);
            const int gainNumerator = (int)(gain * double(gainDenominator));

            qrc = clSetKernelArg(oclKrnBufferGain, arg++, sizeof(dst), &dst);
            if (CL_SUCCESS != qrc)
            {
                std::cerr << "Error setting kernel argument " << (arg - 1) << ".\n";
                return qrc;
            }
        
            qrc = clSetKernelArg(oclKrnBufferGain, arg++, sizeof(src), &src);
            if (CL_SUCCESS != qrc)
            {
                std::cerr << "Error setting kernel argument " << (arg - 1) << ".\n";
                return qrc;
            }
        
            qrc = clSetKernelArg(oclKrnBufferGain, arg++, sizeof(gainNumerator), &gainNumerator);
            if (CL_SUCCESS != qrc)
            {
                std::cerr << "Error setting kernel argument " << (arg - 1) << ".\n";
                return qrc;
            }
        
            qrc = clSetKernelArg(oclKrnBufferGain, arg++, sizeof(gainDenominator), &gainDenominator);
            if (CL_SUCCESS != qrc)
            {
                std::cerr << "Error setting kernel argument " << (arg - 1) << ".\n";
                return qrc;
            }
        
            qrc = clSetKernelArg(oclKrnBufferGain, arg++, sizeof(offset), &offset);
            if (CL_SUCCESS != qrc)
            {
                std::cerr << "Error setting kernel argument " << (arg - 1) << ".\n";
                return qrc;
            }

            size_t globalWorkSize = bufSize / 4 / byteDepth;
            qrc = clEnqueueNDRangeKernel(oclCmdQueue, oclKrnBufferGain, 1, NULL, &globalWorkSize, NULL, waitEvtCnt, waitEvts, kernelEvt);
            if (CL_SUCCESS != qrc)
            {
                std::cerr << "Error queueing the kernel for execution.\n";
                return qrc;
            }

            return CL_SUCCESS;
        };

    //// Run the acquisition and perform processing.

    // Synchronization handle to ensure our helder threads are done before we unwind resources they reference.
    std::atomic<size_t> threadSyncRefCnt (0);
    unwinder.spool([&](){ while (threadSyncRefCnt.load()) std::this_thread::yield(); });

    // Loop until an error, or the user presses a key.

    auto lastCliTime = clock() * 1000 / CLOCKS_PER_SEC;

    // Start BitFlow acquisition.
    rc = BiCirControl(hBoard, &hBiba, BISTART, BiAsync);
    if (BF_OK != rc)
    {
        std::cerr << "Failed to start the BitFlow acquisition engine.\n";
        return EXIT_FAILURE;
    }

#if USE_GL_DISPLAY
    while (!glutHasExited.load())
#else
    std::cout << "\nEntering the acquisition loop. Press any key to exit...\n\n";
    while (!BFkbhit())
#endif
    {
        // Wait for the next acquisition buffer.
        BiCirHandle hBuf;
        rc = BiCirWaitDoneFrame(hBoard, &hBiba, 1000, &hBuf);
        if (BF_OK != rc)
        {
            std::cerr << "An error occurred waiting for the next frame.\n";
            return EXIT_FAILURE;
        }

        const auto timeNow = clock() * 1000 / CLOCKS_PER_SEC;
        if (timeNow - lastCliTime >= 1000 / 60)
        {
            lastCliTime = timeNow;

            std::cout.width(0);
            std::cout.fill(' ');
            std::cout << "\rFrame Count: " << std::dec << hBuf.FrameCount;

#if !USE_GL_DISPLAY
            std::cout << "\tFancy Display: ";
            std::cout.width(8);
            std::cout.fill('0');
            std::cout << std::hex << *reinterpret_cast<PBFU32>(sysDispBuf.data()) << "\t";
#endif

            std::cout.flush();
        }

        // Hold the buffer until we're done with it.
        // Make the acquisition buffer available again.
        rc = BiCirStatusSet(hBoard, &hBiba, hBuf, BIHOLD);
        if (BF_OK != rc)
        {
            std::cerr << "Error setting the buffer status to BIHOLD.\n";
            return EXIT_FAILURE;
        }

        // Select the current OpenCL buffer.
        const auto oclProcBuf = oclBufs[hBuf.BufferNumber];

        // Enqueue the kernel operation.
        clrc = enqueue_buffer_gain(oclProcBuf, oclProcBuf, imageByteSize, imageByteDepth, oclGain, oclOffset, 0, NULL, NULL);
        if (CL_SUCCESS != clrc)
        {
            std::cerr << "Failed to enqueue the kernel operation.\n";
            return EXIT_FAILURE;
        }

        // An even triggered when we're done with the OpenCL buffer.
        cl_event bufferAvailEvent;

#if USE_GL_DISPLAY
        // Display the resultant buffer in the GLUT display.
        clrc = enqueueSetMainBFsimpleGLUT(oclCmdQueue, oclProcBuf, 0, NULL, &bufferAvailEvent);
        if (CL_SUCCESS != clrc)
        {
            std::cerr << "Failed to enqueue the buffer display operation.\n";
            return EXIT_FAILURE;
        }
#else
        // Enqueue the device-to-host buffer write operation.
        clrc = clEnqueueReadBuffer(oclCmdQueue, oclProcBuf, CL_NON_BLOCKING, 0, imageByteSize, sysDispBuf.data(), NULL, NULL, &bufferAvailEvent);
        if (CL_SUCCESS != clrc)
        {
            std::cerr << "Failed to enqueue the host-to-device buffer write operation.\n";
            return EXIT_FAILURE;
        }
#endif

        // Concurrently...
        threadSyncRefCnt.fetch_add(1);
        std::async([&threadSyncRefCnt, bufferAvailEvent, hBoard, &hBiba, hBuf]()
        {
            // Wait for the buffer copy to complete.
            if (CL_SUCCESS != clWaitForEvents(1, &bufferAvailEvent))
                std::cerr << "Error waiting for the buffer write event to complete.\n";

            // Release the event.
            if (CL_SUCCESS != clReleaseEvent(bufferAvailEvent))
                std::cerr << "Error releasing the buffer write event.\n";

            // Make the BitFlow buffer available for acquisition again.
            if (BF_OK != BiCirStatusSet(hBoard, &hBiba, hBuf, BIAVAILABLE))
                std::cerr << "Error setting the buffer status to BIAVAILABLE.\n";

            // We're done with our dependency handles.
            threadSyncRefCnt.fetch_sub(1);
        });

        // Flush the OpenCL command queue.
        clrc = clFlush(oclCmdQueue);
        if (CL_SUCCESS != clrc)
        {
            std::cerr << "Failed to flush the OpenCL command queue.\n";
            return EXIT_FAILURE;
        }
    }

    std::cout << "\n\nCleaning up...\n";

    return EXIT_SUCCESS;
}
